/* PROGRAM:	eggsh
 * FILE:	$Header: /home/egg/src/RCS/reg_araneys.c,v 0.9 2011/1/10 22:04:39 ghn Exp $
 * PURPOSE:	Hardware iterface for Araneus REG (USB)
 * AUTHOR:	Boris Fain
 * DATE:	2011-01-10
 *
 * Copyright 2011 Institute of Heartmath
 * Redistributable under the terms of the GNU Public Licence (GPL)
 * 
 */

/* Define this to dump samples from the REG which we actually use into
   a file named dumpreg.dat.  This can be used to debug serial port
   mode problems (for example, setting the character frame to 7 bits,
   or telling it to ignore breaks, which will lose all zero bytes. */
/* #define DUMPREG */

#include <stdio.h>
#ifdef __USE_BSD
#undef __USE_BSD
#endif
#define __USE_BSD
#include <termios.h>
#undef __USE_BSD
#include <unistd.h>
#include <fcntl.h>
#define __USE_BSD
#include <errno.h>
#undef __USE_BSD

#include "global.h"
#include "genlib.h"
#include "reg.h"

/*these includes are for araneus - move to global. */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <fcntl.h>

#include <usb.h>

#define USB_VENDOR_ARANEUS 		0x12d8
#define USB_ARANEUS_PRODUCT_ALEA 	0x0001

/* Number of bytes to read at a time.  Must be a multiple of 64. */
#define BUFFER_SIZE 8192
#define THROW_AWAY_BUFFER 1024

/* Timeout in milliseconds */
#define TIMEOUT 5000

/****************************/

struct usb_dev_handle *handle;
/*this is an ugly declaration, but it allows us to preserve the overall distribution */

#define MAXDEV	20

static int32 bitsleft[MAXDEV];

#ifdef DUMPREG
static FILE *dumpfile;		      /* REG dump file handle */
static unsigned char dumpbuf[1024];   /* REG dump buffer */
static int dumpptr = 0; 	      /* Pointer into dump buffer */
#endif

static void
fatal(char *format, ...) {
    va_list args;
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    fprintf(stderr, "\n");
    exit(1);
}

static int OpenDev(DevOpts *opts) { /*this is opening the device function */
  
speed_t baudcon;

#ifdef DUMPREG
  dumpfile = fopen("dumpreg.dat", "w");
  setbuf(dumpfile, NULL);
  dumpptr = 0;
#endif



/*an Araneus block*/
    int r;
    struct usb_bus *bus;
    struct usb_device *dev;
    


    usb_init();
    usb_find_busses();
    usb_find_devices();

    for (bus = usb_busses; bus != NULL; bus = bus->next) {
	for (dev = bus->devices; dev != NULL; dev = dev->next) {
	    if (dev->descriptor.idVendor == USB_VENDOR_ARANEUS &&
		dev->descriptor.idProduct == USB_ARANEUS_PRODUCT_ALEA)
		goto found;
	}
    }
    fatal("No Alea device found\n");

 found:
    handle = usb_open(dev);
    r = usb_claim_interface(handle, 0);
    if (r < 0)
	fatal("unable to claim Alea interface 0");


/***********************/



  switch(opts->baud) { /*this may be handy for passing options to this thing!*/
  case 1200: baudcon = B1200; break;
  case 2400: baudcon = B2400; break;
  case 4800: baudcon = B4800; break;
  case 9600: baudcon = B9600; break;
  case 19200: baudcon = B19200; break;
  case 38400: baudcon = B38400; break;
  default:
    printf("%s: Baud rate %ld not supported.\n", pgmname, opts->baud);
    /*return -1; */
  }
  
  usb_release_interface(handle, 0);
  /*usb_close(handle); */ /* the handle will be closed further on */
  
  return r;  /* that's all we really need, but actually we should re-initialize every time? */
}

static int32 Sample(int32 dd, uint16 bits) {
	
	
  /*ARANEUS*********/
    int nbytes = bits; 	/* Number of bytes to read, or -1 if no limit */
    int r, tmp1, tmp2;
    char buffer[BUFFER_SIZE];
    unsigned int remaining;
    int32 sum, sum_xor;

    r = usb_claim_interface(handle, 0);
    if (r < 0) {
    	return -1;
		fatal("unable to claim Alea interface 0");
    }

    remaining = nbytes == -1 ? BUFFER_SIZE : nbytes;
    while (remaining > 0) {
	unsigned long nreq;
	unsigned long nreq_rounded, nreq_rounded_total;
	int i;
	
	nreq = remaining <= BUFFER_SIZE ? remaining : BUFFER_SIZE;
	/* Round the size up to the nearest multiple of 64 bytes */
	nreq_rounded = (nreq + 63) & ~63;
	/* but also, we need to throw away 1024 bytes, so */
	nreq_rounded_total = nreq_rounded + THROW_AWAY_BUFFER;
	unsigned long nread;
	nread = usb_bulk_read(handle, 1, buffer, nreq_rounded_total, TIMEOUT);
	/*we need to get 100 bits, which is 25 bytes, but then we need to convert*/
	/*them to binary... so we need to convert each of these, then sum them up*/
	if (nread < 0) {
		return -1;
	    fatal("error reading from Alea device");
	}
	if (nread != nreq_rounded_total) {
		return -1;
	    fatal("short read reading from Alea device");
	}
	/*now the request is to XOR the bits with 01010101b 
	 * to get rid of bias, and we can do this while sampling
	 * but let's try to do two sums to check whether this is OK */
	    
	sum = sum_xor = 0;
	int even = 0;
	/*and now we just read the last nreq bytes */
	for (i = nreq_rounded_total-nreq; i < nreq_rounded_total; i++) {
	    if ((int)buffer[i] > -1) {
	    	sum = sum + 1; 
	    	tmp1 = 1;
	    }
	    else
	    	tmp1 = 0; /*so that's the bit*/
	    	
	    /* here we do a real xor
	     * let's do a loop since it's more understandable  */
	    if (even) {
	    	if (tmp1) tmp2 = 1;
	    	else tmp2 = 0; 
	    }
	    else {
	    	if (tmp1) tmp2 = 0;
	    	else tmp2 = 1;
	    }
	    
	    sum_xor = sum_xor + tmp2;
	    if (even)
	    	even = 0;
	    else
	        even = 1;
	}
	#ifndef NO_UI
		printf("\nsum = %i, sum_xor = %i, %i", (int)sum, (int)sum_xor, i);
	#endif
	/*printf("\nsum = %i", sum, i);*/
	if (nbytes != -1)
	    remaining -= nreq;
    }
    usb_release_interface(handle, 0);
    /*usb_close(handle);*/ 
	/*in fact do we ever close the handle? Only when close device. */
  return sum_xor;
}

#define SAMP_PERIOD	1000	/* msec */
#define MARGIN		.95	/* how much to headroom to allow in
				   speed measurement */

static int32 EvalSpeed(int32 dd) { /*and here we would see if we can get the required bits in 1 second */
  struct timeval start, end;
  int32 bitct, samp;

  if (dd < 0) return -1;

  gettimeofday(&start, NULL);
  bitct = 0;
  while (1) {
    gettimeofday(&end, NULL);
    if (deltams(&end, &start) >= SAMP_PERIOD) break;
    samp = Sample(dd, 1);
    bitct++;
  }

  return (int32)(bitct * MARGIN);
}

static int32 Discard(int32 dd) {
  int32 disc;
#ifdef CPU_BOUND
  int32  n1;
  uint8 c1;
#endif
  
  if (dd < 0) return -1;

  disc = bitsleft[dd];
  bitsleft[dd] = 0;

#ifdef CPU_BOUND
  do {
    n1 = read(dd, &c1, 1);
    if (n1 == 1) disc++;
  } while (n1 == 1);
#else
  tcflush(dd, TCIFLUSH);
#endif

  return disc;
}

static int32 CloseDev(int32 dd) {
  if (dd < 0) return -1;
  usb_close(handle);
  close(dd);
  return 0;
}

/*  Driver description table.  */

REG_driver REG_araneus = {
                        "ARANEUS", 
			0,
			OpenDev,
			EvalSpeed,
			Sample,
			Discard,
			CloseDev
		      };
